"""empty message

Revision ID: 9d566d915a2c
Revises: 
Create Date: 2022-09-04 19:01:51.002043

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9d566d915a2c'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('doi_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_doi_identifiers_id'), 'doi_identifiers', ['id'], unique=False)
    op.create_table('ensembl_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_ensembl_identifiers_id'), 'ensembl_identifiers', ['id'], unique=False)
    op.create_table('genome_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_genome_identifiers_id'), 'genome_identifiers', ['id'], unique=False)
    op.create_table('keywords',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('text', sa.String(), nullable=False),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('text')
    )
    op.create_index(op.f('ix_keywords_id'), 'keywords', ['id'], unique=False)
    op.create_table('pubmed_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('reference_html', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_pubmed_identifiers_id'), 'pubmed_identifiers', ['id'], unique=False)
    op.create_table('refseq_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('reference_html', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_refseq_identifiers_id'), 'refseq_identifiers', ['id'], unique=False)
    op.create_table('roles',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_roles_name'), 'roles', ['name'], unique=False)
    op.create_table('sra_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_sra_identifiers_id'), 'sra_identifiers', ['id'], unique=False)
    op.create_table('uniprot_identifiers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('identifier', sa.String(), nullable=False),
    sa.Column('db_name', sa.String(), nullable=False),
    sa.Column('db_version', sa.String(), nullable=True),
    sa.Column('url', sa.String(), nullable=True),
    sa.Column('reference_html', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_uniprot_identifiers_id'), 'uniprot_identifiers', ['id'], unique=False)
    op.create_table('users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('username', sa.String(), nullable=False),
    sa.Column('first_name', sa.String(), nullable=True),
    sa.Column('last_name', sa.String(), nullable=True),
    sa.Column('is_superuser', sa.Boolean(), nullable=False),
    sa.Column('is_staff', sa.Boolean(), nullable=False),
    sa.Column('is_active', sa.Boolean(), nullable=False),
    sa.Column('date_joined', sa.DateTime(), nullable=True),
    sa.Column('email', sa.String(), nullable=True),
    sa.Column('last_login', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_users_username'), 'users', ['username'], unique=False)
    op.create_table('wild_type_sequences',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sequence_type', sa.String(), nullable=False),
    sa.Column('sequence', sa.String(), nullable=False),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_wild_type_sequences_id'), 'wild_type_sequences', ['id'], unique=False)
    op.create_table('access_keys',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('key_id', sa.String(), nullable=False),
    sa.Column('public_key', sa.String(), nullable=False),
    sa.Column('name', sa.String(), nullable=True),
    sa.Column('expiration_date', sa.Date(), nullable=True),
    sa.Column('creation_time', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_access_keys_key_id'), 'access_keys', ['key_id'], unique=True)
    op.create_table('experiment_sets',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('urn', sa.String(length=64), nullable=True),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('method_text', sa.String(), nullable=False),
    sa.Column('abstract_text', sa.String(), nullable=False),
    sa.Column('short_description', sa.String(), nullable=False),
    sa.Column('extra_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    sa.Column('private', sa.Boolean(), nullable=False),
    sa.Column('approved', sa.Boolean(), nullable=False),
    sa.Column('published_date', sa.Date(), nullable=True),
    sa.Column('processing_state', sa.String(), nullable=True),
    sa.Column('num_experiments', sa.Integer(), nullable=False),
    sa.Column('created_by_id', sa.Integer(), nullable=True),
    sa.Column('modified_by_id', sa.Integer(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['created_by_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['modified_by_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_experiment_sets_id'), 'experiment_sets', ['id'], unique=False)
    op.create_table('reference_genomes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('short_name', sa.String(), nullable=False),
    sa.Column('organism_name', sa.String(), nullable=False),
    sa.Column('genome_identifier_id', sa.Integer(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['genome_identifier_id'], ['genome_identifiers.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_reference_genomes_id'), 'reference_genomes', ['id'], unique=False)
    op.create_table('users_roles',
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('role_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['role_id'], ['roles.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('user_id', 'role_id')
    )
    op.create_table('experiment_set_doi_identifiers',
    sa.Column('experiment_set_id', sa.Integer(), nullable=False),
    sa.Column('doi_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['doi_identifier_id'], ['doi_identifiers.id'], ),
    sa.ForeignKeyConstraint(['experiment_set_id'], ['experiment_sets.id'], ),
    sa.PrimaryKeyConstraint('experiment_set_id', 'doi_identifier_id')
    )
    op.create_table('experiment_set_keywords',
    sa.Column('experiment_set_id', sa.Integer(), nullable=False),
    sa.Column('keyword_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_set_id'], ['experiment_sets.id'], ),
    sa.ForeignKeyConstraint(['keyword_id'], ['keywords.id'], ),
    sa.PrimaryKeyConstraint('experiment_set_id', 'keyword_id')
    )
    op.create_table('experiment_set_pubmed_identifiers',
    sa.Column('experiment_set_id', sa.Integer(), nullable=False),
    sa.Column('pubmed_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_set_id'], ['experiment_sets.id'], ),
    sa.ForeignKeyConstraint(['pubmed_identifier_id'], ['pubmed_identifiers.id'], ),
    sa.PrimaryKeyConstraint('experiment_set_id', 'pubmed_identifier_id')
    )
    op.create_table('experiment_set_sra_identifiers',
    sa.Column('experiment_set_id', sa.Integer(), nullable=False),
    sa.Column('sra_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_set_id'], ['experiment_sets.id'], ),
    sa.ForeignKeyConstraint(['sra_identifier_id'], ['sra_identifiers.id'], ),
    sa.PrimaryKeyConstraint('experiment_set_id', 'sra_identifier_id')
    )
    op.create_table('experiments',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('urn', sa.String(length=64), nullable=True),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('short_description', sa.String(), nullable=False),
    sa.Column('abstract_text', sa.String(), nullable=False),
    sa.Column('method_text', sa.String(), nullable=False),
    sa.Column('extra_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    sa.Column('private', sa.Boolean(), nullable=False),
    sa.Column('approved', sa.Boolean(), nullable=False),
    sa.Column('published_date', sa.Date(), nullable=True),
    sa.Column('processing_state', sa.String(), nullable=True),
    sa.Column('num_scoresets', sa.Integer(), nullable=False),
    sa.Column('experiment_set_id', sa.Integer(), nullable=True),
    sa.Column('created_by_id', sa.Integer(), nullable=True),
    sa.Column('modified_by_id', sa.Integer(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['created_by_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['experiment_set_id'], ['experiment_sets.id'], ),
    sa.ForeignKeyConstraint(['modified_by_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_experiments_id'), 'experiments', ['id'], unique=False)
    op.create_table('experiment_doi_identifiers',
    sa.Column('experiment_id', sa.Integer(), nullable=False),
    sa.Column('doi_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['doi_identifier_id'], ['doi_identifiers.id'], ),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiments.id'], ),
    sa.PrimaryKeyConstraint('experiment_id', 'doi_identifier_id')
    )
    op.create_table('experiment_keywords',
    sa.Column('experiment_id', sa.Integer(), nullable=False),
    sa.Column('keyword_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiments.id'], ),
    sa.ForeignKeyConstraint(['keyword_id'], ['keywords.id'], ),
    sa.PrimaryKeyConstraint('experiment_id', 'keyword_id')
    )
    op.create_table('experiment_pubmed_identifiers',
    sa.Column('experiment_id', sa.Integer(), nullable=False),
    sa.Column('pubmed_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiments.id'], ),
    sa.ForeignKeyConstraint(['pubmed_identifier_id'], ['pubmed_identifiers.id'], ),
    sa.PrimaryKeyConstraint('experiment_id', 'pubmed_identifier_id')
    )
    op.create_table('experiment_sra_identifiers',
    sa.Column('experiment_id', sa.Integer(), nullable=False),
    sa.Column('sra_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiments.id'], ),
    sa.ForeignKeyConstraint(['sra_identifier_id'], ['sra_identifiers.id'], ),
    sa.PrimaryKeyConstraint('experiment_id', 'sra_identifier_id')
    )
    op.create_table('scoresets',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('urn', sa.String(length=64), nullable=True),
    sa.Column('title', sa.String(), nullable=False),
    sa.Column('method_text', sa.String(), nullable=False),
    sa.Column('abstract_text', sa.String(), nullable=False),
    sa.Column('short_description', sa.String(), nullable=False),
    sa.Column('extra_metadata', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    sa.Column('dataset_columns', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    sa.Column('normalised', sa.Boolean(), nullable=False),
    sa.Column('private', sa.Boolean(), nullable=False),
    sa.Column('approved', sa.Boolean(), nullable=False),
    sa.Column('published_date', sa.Date(), nullable=True),
    sa.Column('processing_state', sa.Enum('incomplete', 'processing', 'failed', 'success', name='processingstate', native_enum=False, create_constraint=True, length=32), nullable=True),
    sa.Column('data_usage_policy', sa.String(), nullable=True),
    sa.Column('num_variants', sa.Integer(), nullable=False),
    sa.Column('experiment_id', sa.Integer(), nullable=False),
    sa.Column('licence_id', sa.Integer(), nullable=True),
    sa.Column('replaces_id', sa.Integer(), nullable=True),
    sa.Column('created_by_id', sa.Integer(), nullable=True),
    sa.Column('modified_by_id', sa.Integer(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['created_by_id'], ['users.id'], ),
    sa.ForeignKeyConstraint(['experiment_id'], ['experiments.id'], ),
    sa.ForeignKeyConstraint(['modified_by_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_scoresets_id'), 'scoresets', ['id'], unique=False)
    op.create_table('scoreset_doi_identifiers',
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('doi_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['doi_identifier_id'], ['doi_identifiers.id'], ),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.PrimaryKeyConstraint('scoreset_id', 'doi_identifier_id')
    )
    op.create_table('scoreset_keywords',
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('keyword_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['keyword_id'], ['keywords.id'], ),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.PrimaryKeyConstraint('scoreset_id', 'keyword_id')
    )
    op.create_table('scoreset_meta_analysis_sources',
    sa.Column('source_scoreset_id', sa.Integer(), nullable=False),
    sa.Column('meta_analysis_scoreset_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['meta_analysis_scoreset_id'], ['scoresets.id'], ),
    sa.ForeignKeyConstraint(['source_scoreset_id'], ['scoresets.id'], ),
    sa.PrimaryKeyConstraint('source_scoreset_id', 'meta_analysis_scoreset_id')
    )
    op.create_table('scoreset_pubmed_identifiers',
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('pubmed_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['pubmed_identifier_id'], ['pubmed_identifiers.id'], ),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.PrimaryKeyConstraint('scoreset_id', 'pubmed_identifier_id')
    )
    op.create_table('scoreset_sra_identifiers',
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('sra_identifier_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.ForeignKeyConstraint(['sra_identifier_id'], ['sra_identifiers.id'], ),
    sa.PrimaryKeyConstraint('scoreset_id', 'sra_identifier_id')
    )
    op.create_table('target_genes',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('category', sa.String(), nullable=False),
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('wt_sequence_id', sa.Integer(), nullable=False),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.ForeignKeyConstraint(['wt_sequence_id'], ['wild_type_sequences.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_target_genes_id'), 'target_genes', ['id'], unique=False)
    op.create_table('variants',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('urn', sa.String(length=64), nullable=True),
    sa.Column('data', postgresql.JSONB(astext_type=sa.Text()), nullable=False),
    sa.Column('scoreset_id', sa.Integer(), nullable=False),
    sa.Column('hgvs_nt', sa.String(), nullable=True),
    sa.Column('hgvs_pro', sa.String(), nullable=True),
    sa.Column('hgvs_splice', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['scoreset_id'], ['scoresets.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_variants_id'), 'variants', ['id'], unique=False)
    op.create_table('ensembl_offsets',
    sa.Column('identifier_id', sa.Integer(), nullable=False),
    sa.Column('target_gene_id', sa.Integer(), nullable=False),
    sa.Column('offset', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['identifier_id'], ['ensembl_identifiers.id'], ),
    sa.ForeignKeyConstraint(['target_gene_id'], ['target_genes.id'], ),
    sa.PrimaryKeyConstraint('identifier_id', 'target_gene_id')
    )
    op.create_table('reference_maps',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('is_primary', sa.Boolean(), nullable=False),
    sa.Column('genome_id', sa.Integer(), nullable=False),
    sa.Column('target_id', sa.Integer(), nullable=False),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['genome_id'], ['reference_genomes.id'], ),
    sa.ForeignKeyConstraint(['target_id'], ['target_genes.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_reference_maps_id'), 'reference_maps', ['id'], unique=False)
    op.create_table('refseq_offsets',
    sa.Column('identifier_id', sa.Integer(), nullable=False),
    sa.Column('target_gene_id', sa.Integer(), nullable=False),
    sa.Column('offset', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['identifier_id'], ['refseq_identifiers.id'], ),
    sa.ForeignKeyConstraint(['target_gene_id'], ['target_genes.id'], ),
    sa.PrimaryKeyConstraint('identifier_id', 'target_gene_id')
    )
    op.create_table('uniprot_offsets',
    sa.Column('identifier_id', sa.Integer(), nullable=False),
    sa.Column('target_gene_id', sa.Integer(), nullable=False),
    sa.Column('offset', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['identifier_id'], ['uniprot_identifiers.id'], ),
    sa.ForeignKeyConstraint(['target_gene_id'], ['target_genes.id'], ),
    sa.PrimaryKeyConstraint('identifier_id', 'target_gene_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('uniprot_offsets')
    op.drop_table('refseq_offsets')
    op.drop_index(op.f('ix_reference_maps_id'), table_name='reference_maps')
    op.drop_table('reference_maps')
    op.drop_table('ensembl_offsets')
    op.drop_index(op.f('ix_variants_id'), table_name='variants')
    op.drop_table('variants')
    op.drop_index(op.f('ix_target_genes_id'), table_name='target_genes')
    op.drop_table('target_genes')
    op.drop_table('scoreset_sra_identifiers')
    op.drop_table('scoreset_pubmed_identifiers')
    op.drop_table('scoreset_meta_analysis_sources')
    op.drop_table('scoreset_keywords')
    op.drop_table('scoreset_doi_identifiers')
    op.drop_index(op.f('ix_scoresets_id'), table_name='scoresets')
    op.drop_table('scoresets')
    op.drop_table('experiment_sra_identifiers')
    op.drop_table('experiment_pubmed_identifiers')
    op.drop_table('experiment_keywords')
    op.drop_table('experiment_doi_identifiers')
    op.drop_index(op.f('ix_experiments_id'), table_name='experiments')
    op.drop_table('experiments')
    op.drop_table('experiment_set_sra_identifiers')
    op.drop_table('experiment_set_pubmed_identifiers')
    op.drop_table('experiment_set_keywords')
    op.drop_table('experiment_set_doi_identifiers')
    op.drop_table('users_roles')
    op.drop_index(op.f('ix_reference_genomes_id'), table_name='reference_genomes')
    op.drop_table('reference_genomes')
    op.drop_index(op.f('ix_experiment_sets_id'), table_name='experiment_sets')
    op.drop_table('experiment_sets')
    op.drop_index(op.f('ix_access_keys_key_id'), table_name='access_keys')
    op.drop_table('access_keys')
    op.drop_index(op.f('ix_wild_type_sequences_id'), table_name='wild_type_sequences')
    op.drop_table('wild_type_sequences')
    op.drop_index(op.f('ix_users_username'), table_name='users')
    op.drop_table('users')
    op.drop_index(op.f('ix_uniprot_identifiers_id'), table_name='uniprot_identifiers')
    op.drop_table('uniprot_identifiers')
    op.drop_index(op.f('ix_sra_identifiers_id'), table_name='sra_identifiers')
    op.drop_table('sra_identifiers')
    op.drop_index(op.f('ix_roles_name'), table_name='roles')
    op.drop_table('roles')
    op.drop_index(op.f('ix_refseq_identifiers_id'), table_name='refseq_identifiers')
    op.drop_table('refseq_identifiers')
    op.drop_index(op.f('ix_pubmed_identifiers_id'), table_name='pubmed_identifiers')
    op.drop_table('pubmed_identifiers')
    op.drop_index(op.f('ix_keywords_id'), table_name='keywords')
    op.drop_table('keywords')
    op.drop_index(op.f('ix_genome_identifiers_id'), table_name='genome_identifiers')
    op.drop_table('genome_identifiers')
    op.drop_index(op.f('ix_ensembl_identifiers_id'), table_name='ensembl_identifiers')
    op.drop_table('ensembl_identifiers')
    op.drop_index(op.f('ix_doi_identifiers_id'), table_name='doi_identifiers')
    op.drop_table('doi_identifiers')
    # ### end Alembic commands ###



from collections import Counter
from typing import Dict

from mavedb.lib.exceptions import ValidationError
from mavedb.lib.mave.constants import (
    VARIANT_SCORE_DATA,
    VARIANT_COUNT_DATA,
    REQUIRED_SCORE_COLUMN,
)


# Only compare the columns name
def validate_columns_match(variant, scoreset) -> None:
    """
    Validate that a child matches parents defined columns to keep
    data in sync.
    """
    try:
        if variant.score_columns != scoreset.score_columns:
            if Counter(variant.score_columns) != Counter(scoreset.score_columns):
                raise ValidationError(
                    f"Variant defines score columns '{variant.score_columns}' "
                    f"but parent defines columns '{scoreset.score_columns}. "
                )
        if variant.count_columns != scoreset.count_columns:
            # Ignore the column order but comparing the contents. If they have same contents, pass.
            if Counter(variant.count_columns) != Counter(scoreset.count_columns):
                raise ValidationError(
                    f"Variant defines count columns '{variant.count_columns}' "
                    f"but parent defines columns '{scoreset.count_columns}. "
                )
    except KeyError as error:
        raise ValidationError(f"Missing key {str(error)}")


def validate_variant_json(data: Dict[str, Dict]) -> None:
    """
    Checks a given dictionary to ensure that it is suitable to be used
    as the `data` attribute in a :class:`Variant` instance.

    Parameters
    ----------
    data : dict
        Dictionary of keys mapping to a list.
    """
    expected_keys = [VARIANT_SCORE_DATA, VARIANT_COUNT_DATA]
    for key in expected_keys:
        if key not in data.keys():
            raise ValidationError(
                "Missing the required key '%(key)'.",
                params={"data": data, "key": key},
            )

    if REQUIRED_SCORE_COLUMN not in data[VARIANT_SCORE_DATA]:
        raise ValidationError(
            "Missing required column '%(col)s' in variant's score data.",
            params={"col": REQUIRED_SCORE_COLUMN},
        )
    extras = [k for k in data.keys() if k not in set(expected_keys)]
    if len(extras) > 0:
        extras = [k for k in data.keys() if k not in expected_keys]
        raise ValidationError(
            "Encountered unexpected keys '%(extras)s'.",
            params={"extras": extras},
        )

    # Check the correct data types are given.
    for key in expected_keys:
        if not isinstance(data[key], dict):
            type_ = type(data[key]).__name__
            raise ValidationError(
                "Value for '%(key)' must be a dict not %(type)s.",
                params={"key": key, "type": type_},
            )



import json
import logging
import os

import uvicorn
from fastapi import FastAPI
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.utils import get_openapi
from requests import Request
from slack_sdk.webhook import WebhookClient
from sqlalchemy.orm import configure_mappers
from starlette import status
from starlette.responses import JSONResponse, Response

from mavedb.models import *

from mavedb import __version__
from mavedb.routers import (
    access_keys,
    api_information,
    doi_identifiers,
    experiment_sets,
    experiments,
    licenses,
    pubmed_identifiers,
    target_gene_identifiers,
    raw_read_identifiers,
    reference_genomes,
    scoresets,
    target_genes,
    users,
)

logging.basicConfig()
# Un-comment this line to log all database queries:
# logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
logger = logging.getLogger(__name__)


# Scan all our model classes and create backref attributes. Otherwise, these attributes only get added to classes once
# an instance of the related class has been created.
configure_mappers()

app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.include_router(access_keys.router)
app.include_router(api_information.router)
app.include_router(doi_identifiers.router)
app.include_router(experiment_sets.router)
app.include_router(experiments.router)
app.include_router(licenses.router)
app.include_router(pubmed_identifiers.router)
app.include_router(raw_read_identifiers.router)
app.include_router(reference_genomes.router)
app.include_router(scoresets.router)
app.include_router(target_gene_identifiers.router)
app.include_router(target_genes.router)
app.include_router(users.router)


@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    return JSONResponse(
        status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
        content=jsonable_encoder({"detail": list(map(lambda error: customize_validation_error(error), exc.errors()))}),
    )

def customize_validation_error(error):
    if error["type"] == "type_error.none.not_allowed":
        return {"loc": error["loc"], "msg": "Required", "type": error["type"]}
    return error


def exception_as_dict(ex):
    return dict(
        type=ex.__class__.__name__,
        exception=str(ex),
    )


@app.exception_handler(Exception)
async def exception_handler(request, err):
    logger.error("Uncaught exception", exc_info=err)
    slack_webhook_url = os.getenv("SLACK_WEBHOOK_URL")
    if slack_webhook_url is not None and len(slack_webhook_url) > 0:
        client = WebhookClient(url=slack_webhook_url)
        response = client.send(
            text=json.dumps(exception_as_dict(err)),
            blocks=[
                {
                    "type": "section",
                    "text": {
                        "type": "plain_text",
                        "text": json.dumps(exception_as_dict(err)),
                    }
                }
            ]
        )
    return JSONResponse(status_code=500, content={"message": "Internal server error"})


def customize_openapi_schema():
    title = "MaveDB API"
    version = __version__
    openapi_schema = get_openapi(title=title, version=version, routes=app.routes)
    openapi_schema["info"] = {
        "title": title,
        "version": version,
        "description": """MaveDB is a public repository for datasets from Multiplexed Assays of Variant Effect (MAVEs),
such as those generated by deep mutational scanning (DMS) or massively parallel reporter assay (MPRA) experiments.""",
        # 'termsOfService': 'url',
        "contact": {
            "name": "MavaDB/CAVA software group",
            "url": "https://github.com/VariantEffect/mavedb-api/issues",
            "email": "rubin.a@wehi.edu.au",
        },
        "license": {
            "name": "Gnu Affero General Public License 3.0",
            "url": "https://www.gnu.org/licenses/agpl-3.0.en.html",
        },
    }
    app.openapi_schema = openapi_schema
    return app.openapi_schema


customize_openapi_schema()


# If the application is not already being run within a uvicorn server, start uvicorn here.
if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=8000)



from mavedb.view_models.target_gene import TargetGeneCreate

import pytest


def test_create_target_gene(test_empty_db):
    name = "UBE2I"
    category = "Regulatory"
    external_identifiers = [{"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 1}]
    reference_maps = [{"genomeId": 1}]
    wt_sequence = {
        "sequenceType": "dna",
        "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGA"
                    "TGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAA"
                    "TGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAAT"
                    "GACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACAC"
                    "TGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGG"
                    "AACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTA"
                    "CTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTAT"
                    "TGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGA"
                    "CGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
    }
    externalIdentifier = TargetGeneCreate(
        name=name,
        category=category,
        reference_maps=reference_maps,
        external_identifiers=external_identifiers,
        wt_sequence=wt_sequence,
    )
    assert externalIdentifier.name == "UBE2I"
    assert externalIdentifier.category == "Regulatory"


def test_create_invalid_category(test_empty_db):
    name = "UBE2I"
    invalid_category = "invalid name"
    external_identifiers = [{"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0}]
    reference_maps = [{"genomeId": 1}]
    wt_sequence = {
        "sequenceType": "dna",
        "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGA"
                    "TGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAA"
                    "TGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAAT"
                    "GACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACAC"
                    "TGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGG"
                    "AACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTA"
                    "CTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTAT"
                    "TGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGA"
                    "CGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
    }
    with pytest.raises(ValueError) as exc_info:
        TargetGeneCreate(
            name=name,
            category=invalid_category,
            reference_maps=reference_maps,
            external_identifiers=external_identifiers,
            wt_sequence=wt_sequence,
        )
    assert (
            "invalid name is not a valid target category. Valid categories are Protein coding, Regulatory, and Other"
            " noncoding"
            in str(exc_info.value)
    )


def test_create_invalid_sequence_type(test_empty_db):
    name = "UBE2I"
    category = "Regulatory"
    external_identifiers = [{"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0}]
    reference_maps = [{"genomeId": 1}]
    wt_sequence = {
        "sequenceType": "dnaa",
        "sequence": "ATGAGTATTCAACATTTCCGTGTCGCCCTTATTCCCTTTTTTGCGGCATTTTGCCTTCCTGTTTTTGCTCACCCAGAAACGCTGGTGAAAGTAAAAGA"
                    "TGCTGAAGATCAGTTGGGTGCACGAGTGGGTTACATCGAACTGGATCTCAACAGCGGTAAGATCCTTGAGAGTTTTCGCCCCGAAGAACGTTTTCCAA"
                    "TGATGAGCACTTTTAAAGTTCTGCTATGTGGCGCGGTATTATCCCGTGTTGACGCCGGGCAAGAGCAACTCGGTCGCCGCATACACTATTCTCAGAAT"
                    "GACTTGGTTGAGTACTCACCAGTCACAGAAAAGCATCTTACGGATGGCATGACAGTAAGAGAATTATGCAGTGCTGCCATAACCATGAGTGATAACAC"
                    "TGCGGCCAACTTACTTCTGACAACGATCGGAGGACCGAAGGAGCTAACCGCTTTTTTGCACAACATGGGGGATCATGTAACTCGCCTTGATCGTTGGG"
                    "AACCGGAGCTGAATGAAGCCATACCAAACGACGAGCGTGACACCACGATGCCTGCAGCAATGGCAACAACGTTGCGCAAACTATTAACTGGCGAACTA"
                    "CTTACTCTAGCTTCCCGGCAACAATTAATAGACTGGATGGAGGCGGATAAAGTTGCAGGACCACTTCTGCGCTCGGCCCTTCCGGCTGGCTGGTTTAT"
                    "TGCTGATAAATCTGGAGCCGGTGAGCGTGGGTCTCGCGGTATCATTGCAGCACTGGGGCCAGATGGTAAGCCCTCCCGTATCGTAGTTATCTACACGA"
                    "CGGGGAGTCAGGCAACTATGGATGAACGAAATAGACAGATCGCTGAGATAGGTGCCTCACTGATTAAGCATTGGTAA",
    }
    with pytest.raises(ValueError) as exc_info:
        TargetGeneCreate(
            name=name,
            category=category,
            reference_maps=reference_maps,
            external_identifiers=external_identifiers,
            wt_sequence=wt_sequence,
        )
    assert f"'{wt_sequence['sequenceType']}' is not a valid sequence type" in str(exc_info.value)


def test_create_not_match_sequence_and_type(test_empty_db):
    name = "UBE2I"
    category = "Regulatory"
    external_identifiers = [{"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0}]
    reference_maps = [{"genomeId": 1}]
    wt_sequence = {"sequenceType": "dna", "sequence": "ARCG"}
    with pytest.raises(ValueError) as exc_info:
        TargetGeneCreate(
            name=name,
            category=category,
            reference_maps=reference_maps,
            external_identifiers=external_identifiers,
            wt_sequence=wt_sequence,
        )
    assert f"invalid {wt_sequence['sequenceType']} sequence provided" in str(exc_info.value)


def test_create_invalid_sequence(test_empty_db):
    name = "UBE2I"
    category = "Regulatory"
    external_identifiers = [{"identifier": {"dbName": "Ensembl", "identifier": "ENSG00000103275"}, "offset": 0}]
    reference_maps = [{"genomeId": 1}]
    wt_sequence = {"sequenceType": "dna", "sequence": "AOCG%"}
    with pytest.raises(ValueError) as exc_info:
        TargetGeneCreate(
            name=name,
            category=category,
            reference_maps=reference_maps,
            external_identifiers=external_identifiers,
            wt_sequence=wt_sequence,
        )
    assert f"invalid {wt_sequence['sequenceType']} sequence provided" in str(exc_info.value)



from sqlalchemy import Column, Integer, String, SmallInteger, TIMESTAMP

from mavedb.db.base import Base


# TODO Unused
"""
class FailedTask(Base):
    # __tablename__ = 'core_failedtask'
    __tablename__ = 'failed_tasks'

    id = Column(Integer, autoincrement=True, primary_key=True)
    # creation_timestamp = Column('creation_date', TIMESTAMP(timezone=True), nullable=False)  # TODO Defaults
    # modification_timestamp = Column('modification_date', TIMESTAMP(timezone=True), nullable=True)  # TODO Defaults
    creation_timestamp = Column(TIMESTAMP(timezone=True), nullable=False)  # TODO Defaults
    modification_timestamp = Column(TIMESTAMP(timezone=True), nullable=True)  # TODO Defaults
    celery_task_id = Column(String(36), nullable=False)
    name = Column(String(125), nullable=False)
    full_name = Column(String, nullable=False)
    args = Column(String, nullable=True)
    kwargs = Column(String, nullable=True)
    exception_class = Column(String, nullable=False)
    # exception_message = Column('exception_msg', String, nullable=False)
    exception_message = Column(String, nullable=False)
    # trace = Column('traceback', String, nullable=True)
    trace = Column(String, nullable=True)
    num_failures = Column(SmallInteger, nullable=False)
    user_id = Column(Integer, nullable=True)  # TODO
"""



from typing import Optional, Literal

from pydantic import Field

from mavedb.view_models.base.base import BaseModel


class UserBase(BaseModel):
    """Base class for user view models."""

    username: str = Field(..., alias="orcidId")
    first_name: Optional[str]
    last_name: Optional[str]

    class Config:
        allow_population_by_field_name = True


class CurrentUserUpdate(BaseModel):
    """View model for updating the current user."""

    email: Optional[str]


class AdminUserUpdate(BaseModel):
    """View model for updating current user, for admin clients."""

    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    roles: list[Literal["admin"]]


class SavedUser(UserBase):
    """Base class for user view models representing saved records."""

    class Config:
        orm_mode = True


class User(SavedUser):
    """User view model containing properties visible to non-admin users."""

    pass


class CurrentUser(SavedUser):
    """User view model for information about the current user."""

    email: Optional[str]
    roles: list[str]


class AdminUser(SavedUser):
    """User view model containing properties to return to admin clients."""

    id: int
    roles: list[str]



import re
from collections import defaultdict
from io import StringIO
from itertools import groupby
from operator import itemgetter
from typing import Union, Optional, Tuple, List, TextIO, BinaryIO, Set, Dict

import pandas as pd
import numpy as np
from mavehgvs import MaveHgvsParseError, Variant
from fqfa.util.translate import translate_dna
from fqfa.util.infer import infer_sequence_type

from . import constants
from .utils import is_csv_null, NULL_VALUES, READABLE_NULL_VALUES
from ..validation.constants.general import null_values_re


class MaveDataset:
    class DatasetType:
        SCORES = "scores"
        COUNTS = "counts"

    class HGVSColumns:
        NUCLEOTIDE: str = constants.HGVS_NT_COLUMN
        TRANSCRIPT: str = constants.HGVS_SPLICE_COLUMN
        PROTEIN: str = constants.HGVS_PRO_COLUMN

        @classmethod
        def options(cls) -> List[str]:
            return [cls.NUCLEOTIDE, cls.TRANSCRIPT, cls.PROTEIN]

    class AdditionalColumns:
        @classmethod
        def options(cls) -> List[str]:
            return []

    # ---------------------- Construction------------------------------------ #
    @classmethod
    def for_scores(cls, file: Union[str, TextIO, BinaryIO]) -> "MaveScoresDataset":
        return cls._for_type(file=file, dataset_type=cls.DatasetType.SCORES)

    @classmethod
    def for_counts(cls, file: Union[str, TextIO, BinaryIO]) -> "MaveCountsDataset":
        return cls._for_type(file=file, dataset_type=cls.DatasetType.COUNTS)

    @classmethod
    def _for_type(
        cls, file: Union[str, TextIO, BinaryIO], dataset_type: str
    ) -> Union["MaveScoresDataset", "MaveCountsDataset"]:

        if isinstance(file, str):
            handle = file
        elif hasattr(file, "read"):
            file_contents = file.read()
            if hasattr(file_contents, "decode"):
                file_contents = file_contents.decode("utf-8")
            file_contents = file_contents.strip()
            handle = StringIO(file_contents)
        else:
            raise TypeError(f"Expected file path or buffer object. " f"Got '{type(file).__name__}'")

        extra_na_values = set(
            list(NULL_VALUES)
            + [str(x).lower() for x in NULL_VALUES]
            + [str(x).upper() for x in NULL_VALUES]
            + [str(x).capitalize() for x in NULL_VALUES]
        )

        df = pd.read_csv(
            filepath_or_buffer=handle,
            sep=",",
            encoding="utf-8",
            quotechar='"',
            comment="#",
            na_values=extra_na_values,
            keep_default_na=True,
            dtype={
                **{c: str for c in cls.HGVSColumns.options()},
                MaveScoresDataset.AdditionalColumns.SCORES: float,
            },
        ).replace(null_values_re, np.NaN)

        if dataset_type == cls.DatasetType.SCORES:
            return MaveScoresDataset(df)
        elif dataset_type == cls.DatasetType.COUNTS:
            return MaveCountsDataset(df)
        else:
            raise ValueError(f"'{dataset_type}' is not a recognised dataset type.")

    # ---------------------- Public ----------------------------------------- #
    @property
    def label(self) -> str:
        return "dataset"

    @property
    def is_valid(self) -> Optional[bool]:
        if self._errors is None:
            return None
        return len(self._errors) == 0

    @property
    def n_errors(self) -> Optional[int]:
        if self._errors is None:
            return None
        return len(self._errors)

    @property
    def errors(self) -> Optional[List[str]]:
        return self._errors

    @property
    def is_empty(self) -> bool:
        return self._df.empty

    @property
    def columns(self) -> List[str]:
        return list(self._df.columns)

    @property
    def hgvs_columns(self) -> List[str]:
        return [c for c in self.columns if c in self.HGVSColumns.options()]

    @property
    def non_hgvs_columns(self) -> List[str]:
        return [c for c in self.columns if c not in self.HGVSColumns.options()]

    @property
    def n_rows(self) -> int:
        return len(self._df)

    @property
    def n_columns(self) -> int:
        return len(self.columns)

    @property
    def index_column(self) -> Optional[str]:
        if self._errors:
            return None
        return self._index_column

    @property
    def index(self) -> Optional[pd.Index]:
        if self._errors:
            return None
        return self._df.index.copy(deep=True)

    def data(self, serializable=False) -> pd.DataFrame:
        """
        Return underlying dataframe object.

        Parameters
        ----------
        serializable: bool
            Replaces `np.NaN` with `None` for JSON compatibility.
        """
        if serializable:
            return self._df.where(cond=pd.notnull(self._df), other=None, inplace=False)
        return self._df.copy(deep=True)

    def match_other(self, other: "MaveDataset") -> Optional[bool]:
        """
        Check that each dataset defined the same variants in each column.

        Parameters
        ----------
        other: MaveDataset
            Validator instance to match against.

        Returns
        -------
        A boolean indicating index match, otherwise `None` if either instance
        is not valid.
        """
        if (not self.is_valid) or (not other.is_valid):
            return None

        if self.index_column != other.index_column:
            return False

        return all(self._df[column].equals(other._df[column]) for column in self.HGVSColumns.options())

    def to_dict(self) -> Dict[str, Dict]:
        """
        Returns underlying dataframe as dictionary in 'records' orientation.
        Keys will be index values and values will be an inner dictionary mapping
        column names to row values for said index.
        """
        # Convert np.NaN values to None for consistency across all columns and
        # for compatibility in PostgresSQL queries. Replaces all values which
        # are considered null by pandas with None by masking pd.notnull cells.
        return self._df.where(cond=pd.notnull(self._df), other=None, inplace=False).to_dict(orient="index")

    def validate(
        self,
        targetseq: Optional[str] = None,
        relaxed_ordering: bool = False,
        allow_index_duplicates: bool = False,
    ) -> "MaveDataset":

        self._errors = []
        self._df.index = pd.RangeIndex(start=0, stop=self.n_rows, step=1)
        self._index_column = None

        self._validate_columns()
        # Only attempt to validate variants if columns are valid
        if not self._errors:
            (
                self._normalize_data()
                ._validate_genomic_variants(targetseq, relaxed_ordering)
                ._validate_transcript_variants(targetseq, relaxed_ordering)
                ._validate_protein_variants(targetseq, relaxed_ordering)
                ._validate_index_column(allow_duplicates=allow_index_duplicates)
            )

        if self.is_empty:
            self._errors.append(
                f"No variants could be parsed from your {self.label} file. " f"Please upload a non-empty file."
            )
            return self

        if not self._errors:
            # Set index last as original index is used when indicating duplicate
            # hgvs string row numbers in the column name used as the index (
            # either hgvs_nt when present or hgvs_pro when hgvs_nt is absent).
            self._df.index = pd.Index(self._df[self.index_column])

        return self

    # ---------------------- Private ---------------------------------------- #
    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        index_column: Optional[str] = None,
        errors: Optional[List[str]] = None,
    ):
        self._df: pd.DataFrame = pd.DataFrame() if df is None else df
        self._index_column = index_column or None
        self._errors = None if errors is None else list(errors)

    def __repr__(self):
        return (
            f"<"
            f"{self.__class__.__name__} "
            f"columns={self.columns} "
            f"index={self.index_column} "
            f"valid={self.is_valid}"
            f">"
        )

    @property
    def _column_order(self) -> Dict[str, int]:
        return defaultdict(
            lambda: 100,
            {
                self.HGVSColumns.NUCLEOTIDE: 0,
                self.HGVSColumns.TRANSCRIPT: 1,
                self.HGVSColumns.PROTEIN: 2,
                **{c: (2 + i) for (i, c) in enumerate(self.AdditionalColumns.options(), start=1)},
            },
        )

    def _validate_columns(self) -> "MaveDataset":
        if self._errors:
            return self

        # Pandas will automatically name blank columns using the pattern below
        unnamed = re.compile(r"^Unnamed: \d+$", flags=re.IGNORECASE)
        columns = self.columns
        if any(is_csv_null(h) or unnamed.match(h) for h in columns):
            self._errors.append(
                f"Column names in your {self.label} file cannot values "
                f"considered null such as the following: "
                f"{', '.join(READABLE_NULL_VALUES)}"
            )

        columns = [c for c in columns if not is_csv_null(c)]
        if len(columns) < 1:
            self._errors.append(
                f"No columns could not be parsed from your {self.label} file. "
                "Make sure columns are comma delimited. Column names with "
                "commas must be escaped by enclosing them in double quotes"
            )

        required = {
            self.HGVSColumns.NUCLEOTIDE,
            self.HGVSColumns.PROTEIN,
        }
        if not (set(columns) & required):
            self._errors.append(
                f"Your {self.label} file must define either a nucleotide "
                f"hgvs column '({self.HGVSColumns.NUCLEOTIDE})' "
                f"or a protein hgvs column '({self.HGVSColumns.PROTEIN})'. "
                f"Columns are case-sensitive and must be comma delimited"
            )

        if not (set(columns) - set(self.HGVSColumns.options())):
            self._errors.append(
                f"Your {self.label} file must define at least one additional "
                f"column different from '{self.HGVSColumns.NUCLEOTIDE}', "
                f"'{self.HGVSColumns.TRANSCRIPT}' and "
                f"'{self.HGVSColumns.PROTEIN}'"
            )

        return self

    def _normalize_data(self) -> "MaveDataset":
        if self._errors:
            return self

        # Initialize missing hgvs columns as empty.
        for c in self.HGVSColumns.options():
            if c not in self.columns:
                self._df[c] = np.NaN

        column_order = self._column_order
        sorted_columns = list(sorted(self.columns, key=lambda x: column_order[x]))

        self._df = self._df[sorted_columns]
        return self

    def _validate_genomic_variants(
        self, targetseq: Optional[str] = None, relaxed_ordering: bool = False
    ) -> "MaveDataset":
        if self._column_is_null(self.HGVSColumns.NUCLEOTIDE):
            return self

        defines_transcript_variants = not self._column_is_null(self.HGVSColumns.TRANSCRIPT)
        validated_variants, prefixes, errors = self._validate_variants(
            column=self.HGVSColumns.NUCLEOTIDE,
            splice_defined=defines_transcript_variants,
            targetseq=targetseq,
            relaxed_ordering=relaxed_ordering,
        )

        if ("c" in prefixes or "n" in prefixes) and "g" in prefixes:
            self._errors.append(
                f"{self.HGVSColumns.NUCLEOTIDE}: Genomic variants "
                f"(prefix 'g.') cannot be mixed with transcript variants "
                f"(prefix 'c.' or 'n.')"
            )

        if prefixes == {"g"} and not defines_transcript_variants:
            self._errors.append(
                f"Transcript variants ('{self.HGVSColumns.TRANSCRIPT}' column) "
                f"are required when specifying genomic variants "
                f"(prefix 'g.' in the 'hgvs_nt' column)"
            )

        self._errors += errors

        if not self._errors:
            self._df[self.HGVSColumns.NUCLEOTIDE] = validated_variants

        self._index_column = self.HGVSColumns.NUCLEOTIDE
        return self

    def _validate_transcript_variants(
        self, targetseq: Optional[str] = None, relaxed_ordering: bool = False
    ) -> "MaveDataset":
        defines_nt = not self._column_is_null(self.HGVSColumns.NUCLEOTIDE)
        defines_tx = not self._column_is_null(self.HGVSColumns.TRANSCRIPT)

        if defines_tx and (not defines_nt):
            self._errors.append(
                f"Genomic variants ('{self.HGVSColumns.NUCLEOTIDE}' column) "
                f"must be defined when specifying transcript "
                f"variants ('{self.HGVSColumns.TRANSCRIPT}' column)"
            )

        if not defines_tx:
            return self

        # Don't validate transcript variants against sequence. Might come
        # back to this later with research into implementing gene models.
        validated_variants, _, errors = self._validate_variants(
            column=self.HGVSColumns.TRANSCRIPT,
            targetseq=None,
            relaxed_ordering=relaxed_ordering,
        )

        self._errors += errors

        if not self._errors:
            self._df[self.HGVSColumns.TRANSCRIPT] = validated_variants

        return self

    def _validate_protein_variants(
        self, targetseq: Optional[str] = None, relaxed_ordering: bool = False
    ) -> "MaveDataset":
        if self._column_is_null(self.HGVSColumns.PROTEIN):
            return self

        defines_nt = not self._column_is_null(self.HGVSColumns.NUCLEOTIDE)
        defines_splice = not self._column_is_null(self.HGVSColumns.TRANSCRIPT)

        if defines_splice:
            protein_seq = None
        else:
            protein_seq = targetseq
            if targetseq and "dna" in infer_sequence_type(targetseq).lower():
                protein_seq, remainder = translate_dna(targetseq)
                if remainder:
                    self._errors.insert(
                        0,
                        "Protein variants could not be validated because the "
                        "length of your target sequence is not a multiple of 3",
                    )

        validated_variants, _, errors = self._validate_variants(
            column=self.HGVSColumns.PROTEIN,
            targetseq=protein_seq,
            relaxed_ordering=relaxed_ordering,
        )

        self._errors += errors

        if not self._errors:
            self._df[self.HGVSColumns.PROTEIN] = validated_variants

        if not defines_nt:
            self._index_column = self.HGVSColumns.PROTEIN

        return self

    def _validate_index_column(self, allow_duplicates: bool = False) -> "MaveDataset":
        if self._errors:
            return self

        if self._index_column is None:
            self._index_column = self.HGVSColumns.NUCLEOTIDE

        if self._column_is_partially_null(self._index_column):
            self._errors.append(
                f"Primary column (inferred as '{self._index_column}') "
                f"cannot contain any null values from "
                f"{', '.join(READABLE_NULL_VALUES)} (case-insensitive)"
            )

        if not allow_duplicates:
            dupes = self._df[self._index_column].duplicated(keep=False)
            if np.any(dupes):
                dup_list = zip(
                    self._df.loc[dupes, self._index_column],
                    dupes.index[dupes],
                )
                dupes_str = ", ".join(
                    f"{v}: {[(g[1] + 1) for g in groups]}"  # get row numbers
                    for (v, groups) in groupby(dup_list, key=itemgetter(0))
                )
                self._errors.append(
                    f"Primary column (inferred as '{self._index_column}') "
                    f"contains duplicate HGVS variants: {dupes_str}"
                )

        return self

    def _validate_variants(
        self,
        column: str,
        splice_defined: Optional[bool] = None,
        targetseq: Optional[str] = None,
        relaxed_ordering: bool = False,
    ) -> Tuple[pd.Series, Set[str], List[str]]:

        prefixes = set()
        errors = []

        def validate_variant(variant: str):
            # TODO: logic mirrors that in validate_hgvs_string, which is kept
            #   as a standalone function for backwards compatibility with
            #   django's model validator field. Merge at some point.

            if is_csv_null(variant):
                return np.NaN
            else:
                try:
                    if variant.lower() == "_sy":
                        errors.append("'_sy' is no longer supported and should be " "replaced by 'p.(=)'")
                        return variant
                    elif variant.lower() == "_wt":
                        errors.append(
                            "'_wt' is no longer supported and should be " "replaced by one of 'g.=', 'c.=' or 'n.='"
                        )
                        return variant

                    validated = Variant(
                        variant,
                        targetseq=targetseq,
                        relaxed_ordering=relaxed_ordering,
                    )
                    prefix = validated.prefix.lower()
                    prefixes.add(prefix)

                    prefix_error = self._validate_variant_prefix_for_column(
                        variant=validated,
                        prefix=validated.prefix,
                        column=column,
                        splice_defined=splice_defined,
                    )
                    if prefix_error:
                        errors.append(prefix_error)

                    return str(validated)

                except MaveHgvsParseError as error:
                    errors.append(f"{variant}: {str(error)}")
                    return np.NaN

        validated_variants = self._df[column].apply(validate_variant)

        return validated_variants, prefixes, errors

    def _column_is_null(self, column) -> bool:
        return len(self._df[self._df[column].isna()]) == len(self._df)

    def _column_is_partially_null(self, column) -> bool:
        return 0 < len(self._df[self._df[column].isna()]) < len(self._df)

    def _column_is_fully_specified(self, column) -> bool:
        return len(self._df[self._df[column].isna()]) == 0

    def _validate_variant_prefix_for_column(
        self, variant: Variant, prefix: str, column: str, splice_defined: bool
    ) -> Optional[str]:
        prefix = prefix.lower()

        if column == self.HGVSColumns.NUCLEOTIDE:
            if splice_defined:
                if prefix not in "g":
                    return (
                        f"{column}: "
                        f"'{variant}' is not a genomic variant "
                        f"(prefix 'g.'). Nucleotide variants must "
                        f"be genomic if transcript variants are "
                        f"also present"
                    )
            else:
                if prefix not in "cn":
                    return (
                        f"{column}: "
                        f"'{variant}' is not a transcript variant. "
                        f"The accepted transcript variant prefixes "
                        f"are 'c.' or 'n.'"
                    )
        elif column == self.HGVSColumns.TRANSCRIPT:
            if prefix not in "cn":
                return (
                    f"{column}: "
                    f"'{variant}' is not a transcript variant. The "
                    f"accepted transcript variant prefixes are "
                    f"'c.' or 'n.'"
                )
        elif column == self.HGVSColumns.PROTEIN:
            if prefix not in "p":
                return (
                    f"{column}: "
                    f"'{variant}' is not a protein variant. "
                    f"The accepted protein variant prefix is 'p.'"
                )
        else:
            raise ValueError(f"Unknown column '{column}'. Expected one " f"of {', '.join(self.HGVSColumns.options())}")

        return None


class MaveScoresDataset(MaveDataset):
    class AdditionalColumns:
        SCORES = constants.REQUIRED_SCORE_COLUMN

        @classmethod
        def options(cls) -> List[str]:
            return [cls.SCORES]

    @property
    def label(self) -> str:
        return "scores"

    def _validate_columns(self) -> "MaveDataset":
        super()._validate_columns()

        if self.AdditionalColumns.SCORES not in self.columns:
            self._errors.append(
                f"Your scores dataset is missing the "
                f"'{self.AdditionalColumns.SCORES}' column. "
                f"Columns are case-sensitive and must be comma delimited"
            )

        return self

    def _normalize_data(self) -> "MaveDataset":
        super()._normalize_data()

        should_be_numeric = [self.AdditionalColumns.SCORES]
        for c in should_be_numeric:
            if c in self.columns:
                try:
                    self._df[c] = self._df[c].astype(dtype=float, errors="raise")
                except ValueError as e:
                    self._errors.append(f"{c}: {str(e)}")

        return self


class MaveCountsDataset(MaveDataset):
    @property
    def label(self) -> str:
        return "counts"



from celery import Celery
from celery.app.task import Task as CeleryTask
import io
import json
from kombu.exceptions import (
    OperationalError,
    ConnectionLimitExceeded,
    ConnectionError,
    LimitExceeded,
    KombuError,
    ContentDisallowed,
    SerializationError,
    DecodeError,
    EncodeError,
)
import logging
import pandas as pd
import traceback

# Create the celery src and get the logger
from mavedb.deps import get_db
from mavedb.models.failed_task import FailedTask

celery_app = Celery("tasks", broker="pyamqp://guest@rabbit//")
logger = logging.getLogger(__name__)


KOMBU_ERRORS = (
    ConnectionError,
    ConnectionLimitExceeded,
    ContentDisallowed,
    DecodeError,
    EncodeError,
    KombuError,
    LimitExceeded,
    OperationalError,
    SerializationError,
)


def dump_df(df, orient="records"):
    handle = io.StringIO()
    df.to_json(handle, orient=orient)
    handle.seek(0)
    return handle.read()


def record_failed_task(celery_task_id, name, exception, args, kwargs, trace, user_id=None):
    db = next(get_db())
    failed_task = FailedTask(
        celery_task_id=celery_task_id,
        name=name.split(".")[-1],
        full_name=name,
        exception_class=exception.__class__.__name__,
        exception_message=str(exception).strip(),
        kwargs=kwargs,
        trace=trace,
        user_id=user_id,
    )
    if args:
        args = [dump_df(i) if isinstance(i, pd.DataFrame) else i for i in args]
        failed_task.args = json.dumps(list(args))
    if kwargs:
        for key, item in kwargs.items():
            if isinstance(item, pd.DataFrame):
                kwargs[key] = dump_df(item)
        failed_task.kwargs = json.dumps(kwargs, sort_keys=True)
    if not find_existing_failed_task(db, celery_task_id):
        db.add(failed_task)
        db.commit()
        db.refresh(failed_task)
    return failed_task


def find_existing_failed_task(db, celery_task_id):
    if celery_task_id and celery_task_id != "-1":
        query = db.query(FailedTask).filter(FailedTask.celery_task_id == celery_task_id)
        return query.first()
    else:
        return None


class Task(CeleryTask):
    """
    Base task that will save the task to the database and log the error.
    """

    def run(self, *args, **kwargs):
        raise NotImplementedError()

    def apply_async(
        self, args=None, kwargs=None, task_id=None, producer=None, link=None, link_error=None, shadow=None, **options
    ):
        logger.info(f"Applying async celery function '{self}'")
        return super().apply_async(
            args=args,
            kwargs=kwargs,
            task_id=task_id,
            producer=producer,
            link=link,
            link_error=link_error,
            shadow=shadow,
            **options,
        )

    def delay(self, *args, **kwargs):
        """
        Star argument version of :meth:`apply_async`.

        Does not support the extra options enabled by :meth:`apply_async`.

        Arguments:
            *args (Any): Positional arguments passed on to the task.
            **kwargs (Any): Keyword arguments passed on to the task.
        Returns:
            celery.result.AsyncResult: Future promise.
        """
        logger.info(f"Applying delayed celery function '{self}'")
        return self.apply_async(args, kwargs)

    @staticmethod
    def get_user(user):
        # if isinstance(user, User):
        #    pass
        # elif isinstance(user, int):
        #    if User.objects.filter(pk=user).count():
        #        user = User.objects.get(pk=user)
        #    else:
        #        user = None
        # elif isinstance(user, str):
        #    if User.objects.filter(username=user).count():
        #        user = User.objects.get(username=user)
        #    else:
        #        user = None
        # else:
        #    user = None
        #
        # return user
        return None

    def on_failure(self, exc, task_id, args, kwargs, einfo, user=None):
        """
        Error handler.

        This is run by the worker when the task fails.

        Parameters:
        ----------
        exc : `Exception`
            The exception raised by the task.
        task_id : `str`
            Unique id of the failed task.
        args : `Tuple`
            Original arguments for the task that failed.
        kwargs : `Dict`
            Original keyword arguments for the task that failed.
        einfo : `~billiard.einfo.ExceptionInfo`
            Exception information.
        user : `User`, `int` or `str`.
            User that called the task. Applicable to `publish_scoreset`
            and `create_variants` tasks. Will search User model if user
            is an int pk or str username.

        Returns
        -------
        `None`
            The return value of this handler is ignored.
        """
        user = self.get_user(user)
        # The kwargs can be potentially big in dataset tasks so truncate
        # the variants key before logging.
        kwargs_str = kwargs.copy()

        # TODO Revisit this.
        variants = str(kwargs_str.get("variants", {}))[0:250]
        if variants in kwargs_str:
            kwargs_str["variants"] = variants

        logger.exception(
            f"{self.name} with id {task_id} called with args={2}, kwargs={kwargs_str}"
            f" raised:\n'{exc}' with traceback:\n{einfo}"
        )
        self.save_failed_task(exc, task_id, args, kwargs, einfo, user)
        super(BaseTask, self).on_failure(exc, task_id, args, kwargs, einfo)

    def save_failed_task(self, exc, task_id, args, kwargs, trace, user=None):
        """
        Save a failed task. If it exists, update the modification_date and failure counter.
        """
        task, _ = FailedTask.update_or_create(
            name=self.name.split(".")[-1],
            full_name=self.name,
            exc=exc,
            task_id=task_id,
            args=args,
            kwargs=kwargs,
            traceback=str(trace).strip(),  # einfo
            user=user,
        )
        return task

    def submit_task(self, args=None, kwargs=None, async_options=None, request=None, countdown=10):
        """
        Calls `task.apply_async` and handles any connection errors by
        logging the error to the `django` default log and saving the
        failed task. If a request object is passed in a warning message will be
        shown to the user using the `messages` contrib module and the task
        will be initialised with the authenticated user as a foreign key.

        Parameters
        ----------
        args : tuple, optional, Default: None
            Un-named task arguments.
        kwargs : dict, optional. Default: None
            Task keyword arguments.
        async_options : dict, optional, Default: None
            Additional kwargs that `apply_async` accepts.
        request : Request, optional. Default: None
            Request object from the view calling this function.
        countdown : int
            Delay before executing celery task.

        Returns
        -------
        tuple[bool, Union[FailedTask, Any]]
            Boolean indicating success or failure, FailedTask or task result.
        """
        if not async_options:
            async_options = {}
        try:
            return (True, self.apply_async(args=args, kwargs=kwargs, countdown=countdown, **async_options))
        except KOMBU_ERRORS as ex:
            logger.exception(
                f"Submitting task {self.name} raised a {ex.__class__.__name__} error. " "Failed task has been saved."
            )
            # TODO If this is triggered by an API request, respond with a warning. Here is how this
            # used to be done in Django:
            # if request:
            #     messages.warning(request, network_message)
            failed_task = record_failed_task(
                celery_task_id="-1",
                name=self.name,
                exception=ex,
                args=args,
                kwargs=kwargs,
                trace=traceback.format_exc(),
                user_id=None,  # None if not request else request.user TODO
            )
            return False, failed_task



from mavedb.view_models.pubmed_identifier import PubmedIdentifierCreate
import pytest


def test_pubmed_identifier_create_validator(test_empty_db):
    # Test valid identifier
    valid_identifier = "20711111"
    pubmed_one = PubmedIdentifierCreate(identifier=valid_identifier)
    assert pubmed_one.identifier == "20711111"


def test_invalid_pubmed_identifier_create_validator(test_empty_db):
    # Test invalid identifier
    invalid_identifier = "not_an_identifier"
    with pytest.raises(ValueError) as exc_info:
        PubmedIdentifierCreate(identifier=invalid_identifier)
    assert "not_an_identifier is not a valid PubMed identifier." in str(exc_info.value)



import idutils

from mavedb.view_models.base.base import BaseModel, validator


class PubmedIdentifierBase(BaseModel):
    identifier: str


class PubmedIdentifierCreate(PubmedIdentifierBase):
    @validator("identifier")
    def must_be_valid_pubmed(cls, v):
        if not idutils.is_pmid(v):
            # ValidationError shows weird error and test can't catch it.
            raise ValueError("{} is not a valid PubMed identifier.".format(v))
        return v


# Properties shared by models stored in DB
class SavedPubmedIdentifier(PubmedIdentifierBase):
    id: int
    url: str
    reference_html: str

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class PubmedIdentifier(SavedPubmedIdentifier):
    pass



__all__ = [
    "access_key",
    "doi_identifier",
    "ensembl_identifier",
    "ensembl_offset",
    "experiment",
    "experiment_set",
    "genome_identifier",
    "keyword",
    "license",
    "pubmed_identifier",
    "raw_read_identifier",
    "reference_genome",
    "reference_map",
    "refseq_identifier",
    "refseq_offset",
    "role",
    "scoreset",
    "target_gene",
    "uniprot_identifier",
    "uniprot_offset",
    "user",
    "variant",
    "wild_type_sequence",
]



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

# TODO Move these to a central config object.
DB_HOST = os.getenv("DB_HOST") or "localhost"
DB_PORT = os.getenv("DB_PORT") or 5432
DB_DATABASE_NAME = os.getenv("DB_DATABASE_NAME")
DB_USERNAME = os.getenv("DB_USERNAME")
DB_PASSWORD = os.getenv("DB_PASSWORD")

# DB_URL = "sqlite:///./sql_app.db"
DB_URL = f"postgresql://{DB_USERNAME}:{DB_PASSWORD}@{DB_HOST}:{DB_PORT}/{DB_DATABASE_NAME}"

engine = create_engine(
    # For PostgreSQL:
    DB_URL
    # For SQLite:
    # DB_URL, connect_args={"check_same_thread": False}
)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
